/*

Keytbl - Change keyboard layout
Copyright (C) 1999 Daniel Augsburger

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


===============================================================================
	(c) 14.11.1999 by Daniel Augsburger
===============================================================================

	Daniel Augsburger                         http://www.jumpgates.com/daniel
	Jasminweg 39                                         daniel@jumpgates.com
	8050 Zrich
	SCHWEIZ                                               ++41 (0)1 312 22 38

===============================================================================

	Description:

		Select any keyboard layout on the fly.
		Displays currently active layout in the menu bar.

===============================================================================
Date					Description													Version
-------------------------------------------------------------------------------
14.11.1999			Erstellung													 0.1
19.11.1999			Suche nach "KEYTABLS.*" in "C:\GEMSYS\KEYTABLS"	 0.101
19.11.1999			Einlesen der Sprache von INF-File					 0.102
21.11.1999			offset vom rechten Bildschirmrand in INF-File	 0.103
21.11.1999	neu:	Sprachen-Auswahl ber Popup							 0.11
24.11.1999			negatives right_offset = Abstand von rechtem
						Bildschirmrand												 0.111
24.11.1999			INF-File abspeichern bei ndern der Sprache		 0.112
24.11.1999			beachtet einen ev. vorhandenen XSSI-Cookie		 0.113
25.11.1999			Routine fr Maus-Click abfangen verbessert		 0.114
19.12.1999	rel:	Version 1.0 (Release 1)									 1.0
30.12.1999			nur noch ein RSC-File fr gr. & kl. Flaggen		 1.001
31.12.1999			Keyboard-Layout wird in SupervisorMode gendert	 1.002
31.12.1999	neu:	mit rechte Moustaste Icon-Set wechseln				 1.01
01.01.2000			fr Kbd-Layout ndern nicht in SupervisorMode	 1.0101
01.01.2000	rel:	Version 1.01 (Release 2)								 1.0101
03.01.2000			get_cookie() aus dali									 1.0102
06.01.2000			Dclosedir() vergessen									 1.0103
07.01.2000	neu:	neues Popup mit eigener Popup-Routine				 1.02
08.01.2000			open_file() korrigiert: Absturz wenn $HOME nicht
						existierte													 1.021
08.01.2000	rel:	Version 1.02 (Release 3)								 1.021
12.02.2000			Kein Absturz mehr, wenn keine Layouts gefunden	 1.022
13.02.2000			Gross-/Klein-Schreibung von Layout-Files egal	 1.023
29.03.2000			Redraw-Bug aus lang_popup() entf. u. aufgerumt	 1.024
01.04.2000	neu:	Icon verschieben mit Alt-Links-Klick				 1.03
01.04.2000	rel:	Version 1.03 (Release 4)								 1.03
15.04.2000	neu:	neue Icon-Verschiebe-Routine: move_position()	 1.04
16.04.2000	neu:	Vertauschen der Tasten mit Scan-Codes 0x29 und
						0x2b mglich (fr Mac- und PC-Tastaturen)			 1.05
20.04.2000			save_cnf() hat keine Parameter mehr, sondern
						greift direkt auf globale Variablen zu				 1.051
23.04.2000			unter GPL gestellt										 1.052
29.04.2000	rel:	Version 1.05 (Release 5)								 1.052
-------------------------------------------------------------------------------
===============================================================================

*/

#undef debug_lang_popup
#undef debug_lang_popup_2
#undef debug_open_file
#undef debug_getlangs
#undef debug_langsel
#undef debug_setlayout
#undef debug_keytbladdr
#undef debug_moveicon

#define keytabls_safecopy



#include <stdlib.h>
#include <stdio.h>

#include <bool.h>
#include <string.h>

#include <dali.h>
#include <toserror.h>
#include <osbind.h>
#include <mintbind.h>
#include <gem.h>

#include "keytbl.h"
#include "main.h"



/*************
   Variables
 *************/

int _AESglobal[15];

int ap_id;
int sm_id;	// ap_id of the screen manager
int work_in[11], work_out[57];
int vdi_handle;
int wcell, hcell, wbox, hbox;

OBJECT *mainmenu;

int quit;

static char keytabls_path[] = "C:\\GEMSYS\\KEYTABLS";
static char keytabls_buf[KEYTABLSLEN];

static char	cnf_buffer[CNF_BUFFER_LEN];		// buffer for INF file
static char cnf_lang_s[4] = "";					// language specification read from INF file

static int noflangs = 0;							// number of languages
static LANGUAGE language[MAXNOOFLANG];			// language information structs
static int noflayouts = 0;							// number of available keyboard layouts
static char layout[MAXNOOFLANG][4];				// suffixes of keytabls files -> indicate language
static int current_lang = 0;						// index in array "language" of current language
static OBJECT sel_tree[OBJ_PER_LANG * (MAXNOOFLANG - 1) + 1];	// object tree for popup
static int menu_x, menu_y;							// position of popup menu

static int right_offset = 20;						// offset from right border of screen
static int top_offset = 1;							// offset from top of screen
static GRECT area;									// active area in menu bar 
static int interval_cnt = 0;

static int icon_set = 0;							// icon set (0: small, 1: big)

static int swap_29_2b = 0;							// if != 0 then swap keys with scan codes 0x29 () and 0x2b ($)

static INFOXSSI *xssi_struct = NULL;			// pointer to xssi struct or NULL

static GRECT screen, desk;
static OBJECT *tree;



/**********
   Macros
 **********/

#define SysBell Cconout(0x7)



/**********************
   Service Procedures
 **********************/

/* returns 1 if point (x/y) is inside g, else 0 */
static int is_inside(int x, int y, GRECT g)
{
	return ((x >= g.g_x) && (x < g.g_x + g.g_w) && (y >= g.g_y) && (y < g.g_y + g.g_h));
}



/*==================================================================================*
 * objc_rect:																								*
 * 	Returns the true screen coordinates an the size of an OBJECT.						*
 *----------------------------------------------------------------------------------*
 * returns: -																								*
 *----------------------------------------------------------------------------------*
 * 	tree:		pointer to object tree																*
 * 	obj:		index of object in tree																*
 * 	rect:		contains the coordinates and size of the object upon return				*
 *==================================================================================*/
static void objc_rect(OBJECT *tree, int obj, GRECT *rect)
{
	objc_offset(tree, obj, &rect->g_x, &rect->g_y);
	rect->g_w = tree[obj].ob_width;
	rect->g_h = tree[obj].ob_height;
}



/*==================================================================================*
 * lang_popup:																								*
 * 	Popupfunktion fr alle TOS/MagiC/N.AES-Versionen. Angepasst fr das Popup		*
 * 	von keytbl.																							*
 *----------------------------------------------------------------------------------*
 * returns: Objektnummer des selektierten Objekts bzw. -1									*
 *----------------------------------------------------------------------------------*
 * 	tree:				Zeiger auf den Objektbaum des Popups									*
 * 	obj:				Objektnummer des Popupwurzelobjekts										*
 * 	x_off,y_off:	Gewnschte Koordinaten der oberen linken Ecke des Popups			*
 *							(werden automatisch angepasst, damit das Popup ganz angezeigt	*
 *							wird)																				*
 *==================================================================================*/
static int lang_popup(OBJECT *tree, int obj, int x_off, int y_off)
{
	GRECT		rect, desk, rect1, rect2, rect3;
	int		item, founditem, olditem;
	int		first_child, last_child;
	int		event, ret;
	unsigned int	uret;
	int		bmsk = 0x01;
	int		leave;
	int		mx, my, mb, ks;

	first_child = tree[obj].ob_head;
	last_child  = tree[obj].ob_tail;

	olditem = -1;

	while(last_child != -1)
	{
		olditem = last_child;
		last_child = tree[last_child].ob_tail;
	}

	wind_update(BEG_MCTRL); 							/* Ich will die Kontrolle */

	last_child = olditem;

	tree[ROOT].ob_x = 0;				/* Anpassen des Parent-Objekts */
	tree[ROOT].ob_y = 0;

	tree[obj].ob_x = x_off; 		/* Anpassen der Popup-Box */
	tree[obj].ob_y = y_off;

	wind_get(0,WF_WORKXYWH, &desk.g_x, &desk.g_y, &desk.g_w, &desk.g_h);

	/* Das POPUP darf nicht ausserhalb des Bildschirms erscheinen) */
	if ((tree[obj].ob_x + tree[obj].ob_width) > (desk.g_x + desk.g_w))
		tree[obj].ob_x = (desk.g_x + desk.g_w - tree[obj].ob_width);
	
	if ((tree[obj].ob_y + tree[obj].ob_height) > (desk.g_y + desk.g_h))
		tree[obj].ob_y = (desk.g_y + desk.g_h - tree[obj].ob_height);

	if (tree[obj].ob_x < 0)
		tree[obj].ob_x = desk.g_x;
	if (tree[obj].ob_y < 0)
		tree[obj].ob_y = desk.g_y;

	rect.g_x = tree[obj].ob_x - 8;
	rect.g_y = tree[obj].ob_y - 8;
	rect.g_w = tree[obj].ob_width  + 16;
	rect.g_h = tree[obj].ob_height + 16;

	form_dial(FMD_START,0, 0, 0, 0, rect.g_x, rect.g_y, rect.g_w, rect.g_h);

	graf_mkstate(&mx, &my, &mb, &ks);

	olditem   = -1;
	founditem = item = objc_find(tree, obj, 1, mx, my); 	/* In Men ? */

	if (item != -1)
		if ((tree[item].ob_state & DISABLED) || !(tree[item].ob_flags & SELECTABLE))
			item = -1;

	if (item != -1)
		tree[item].ob_state |= SELECTED;

	objc_draw(tree, obj, MAX_DEPTH, rect.g_x, rect.g_y, rect.g_w, rect.g_h);

	do
	{
		if (founditem != -1)								/* In Meneintrag */
		{
			#ifdef debug_lang_popup
				printf("in menu: %i\r\n", founditem);
			#endif
			leave = 1;
			objc_rect(tree, founditem, &rect1);
		}
		else	/* Auerhalb Pop-Up-Men */
		{
			#ifdef debug_lang_popup
				printf("ausserhalb\r\n");
			#endif
			leave = 0;
			objc_rect(tree, obj, &rect1);
		}

		graf_mkstate(&mx, &my, &mb, &ks);

		event = evnt_multi(MU_BUTTON | MU_M1,
									1, bmsk, ~mb & bmsk,
									leave, rect1.g_x, rect1.g_y, rect1.g_w, rect1.g_h,
									0, 0, 0, 0, 0,
									NULL,
									0,
									&mx, &my, &mb, &ks,
									&uret, &ret);

		olditem = item;

		founditem = item = objc_find(tree, obj, 1, mx, my);

		if (item != -1)
		{
		   if ((tree[item].ob_state & DISABLED) || !(tree[item].ob_flags & SELECTABLE))
				item = -1;
		}

		#ifdef debug_lang_popup_2
			printf("olditem = %i   item = %i\r\n", olditem, item);
		#endif

		if ((olditem != item) && ((abs(olditem - item) != 2) || (olditem == -1) || (item == -1)))
		{
			// -> mouse is above another language
			if (olditem != -1)
			{
/*
				if (tree[olditem].ob_state & SELECTED)
				{
					tree[olditem].ob_state &= ~SELECTED;
					switch (olditem % OBJ_PER_LANG)
					{
						case 1 :
							tree[olditem + 2].ob_state &= ~SELECTED;
							break;
						case 0 :
							tree[olditem - 2].ob_state &= ~SELECTED;
							break;
					}
				}
				else
				{
					tree[olditem].ob_state |= SELECTED;
					switch (olditem % OBJ_PER_LANG)
					{
						case 1 :
							tree[olditem + 2].ob_state |= SELECTED;
							break;
						case 0 :
							tree[olditem - 2].ob_state |= SELECTED;
							break;
					}
				}
*/
				tree[olditem].ob_state &= ~SELECTED;
				switch (olditem % OBJ_PER_LANG)
				{
					case 1 :
						tree[olditem + 2].ob_state &= ~SELECTED;
						break;
					case 0 :
						tree[olditem - 2].ob_state &= ~SELECTED;
						break;
				}
/*
				objc_rect(tree, olditem, &rect2);
				switch (olditem % OBJ_PER_LANG)
				{
					case 1 :
						objc_rect(tree, olditem + 2, &rect3);
						objc_draw(tree, obj, MAX_DEPTH, rect2.g_x, rect2.g_y, rect2.g_w + rect3.g_w, rect2.g_h + rect3.g_h);
						break;
					case 0 :
						objc_rect(tree, olditem - 2, &rect3);
						objc_draw(tree, obj, MAX_DEPTH, rect3.g_x, rect3.g_y, rect2.g_w + rect3.g_w, rect2.g_h + rect3.g_h);
						break;
				}
				objc_draw(tree, obj, MAX_DEPTH, rect3.g_x, rect3.g_y, rect2.g_w + rect3.g_w, rect2.g_h + rect3.g_h);
*/
				switch (olditem % OBJ_PER_LANG)
				{
					case 1 :
						objc_rect(tree, olditem, &rect2);
						objc_rect(tree, olditem + 2, &rect3);
						break;
					case 0 :
						objc_rect(tree, olditem - 2, &rect2);
						objc_rect(tree, olditem, &rect3);
						break;
				}
				objc_draw(tree, obj, MAX_DEPTH, rect2.g_x, rect2.g_y, rect2.g_w + rect3.g_w, rect2.g_h);
			}

			if (item != -1)
			{
/*
				if (tree[item].ob_state & SELECTED)
				{
					tree[item].ob_state &= ~SELECTED;
					switch (olditem % OBJ_PER_LANG)
					{
						case 1 :
							tree[item + 2].ob_state &= ~SELECTED;
							break;
						case 0 :
							tree[item - 2].ob_state &= ~SELECTED;
							break;
					}
				}
				else
				{
					tree[item].ob_state |= SELECTED;
					switch (item % OBJ_PER_LANG)
					{
						case 1 :
							tree[item + 2].ob_state |= SELECTED;
							break;
						case 0 :
							tree[item - 2].ob_state |= SELECTED;
							break;
					}
				}
*/
				tree[item].ob_state |= SELECTED;
				switch (item % OBJ_PER_LANG)
				{
					case 1 :
						tree[item + 2].ob_state |= SELECTED;
						break;
					case 0 :
						tree[item - 2].ob_state |= SELECTED;
						break;
				}
/*
				objc_rect(tree, item, &rect2);
				switch (item % OBJ_PER_LANG)
				{
					case 1 :
						objc_rect(tree, item + 2, &rect3);
						objc_draw(tree, obj, MAX_DEPTH, rect2.g_x, rect2.g_y, rect2.g_w + rect3.g_w, rect2.g_h + rect3.g_h);
						break;
					case 0 :
						objc_rect(tree, item - 2, &rect3);
						objc_draw(tree, obj, MAX_DEPTH, rect3.g_x, rect3.g_y, rect2.g_w + rect3.g_w, rect2.g_h + rect3.g_h);
						break;
				}
*/
				switch (item % OBJ_PER_LANG)
				{
					case 1 :
						objc_rect(tree, item, &rect2);
						objc_rect(tree, item + 2, &rect3);
						break;
					case 0 :
						objc_rect(tree, item - 2, &rect2);
						objc_rect(tree, item, &rect3);
						break;
				}
				objc_draw(tree, obj, MAX_DEPTH, rect2.g_x, rect2.g_y, rect2.g_w + rect3.g_w, rect2.g_h);
			}
		}
	}	while (!(event & MU_BUTTON));

	form_dial(FMD_FINISH, 0, 0, 0, 0, rect.g_x, rect.g_y, rect.g_w, rect.g_h);

	if (item != -1)
	{
		tree[item].ob_state &= ~SELECTED;
		switch (item % OBJ_PER_LANG)
		{
			case 1 :
				tree[item + 2].ob_state &= ~SELECTED;
				break;
			case 0 :
				tree[item - 2].ob_state &= ~SELECTED;
				break;
		}
	}

	if (mb & bmsk)
		evnt_button(1, bmsk, 0x00, &mx, &my, &mb, &ks);		/* Warte auf Mausknopf */

	wind_update(END_MCTRL);

	return(item);
}



/*************************
   Config file functions
 *************************/

/*==================================================================================*
 * open_file:																								*
 * 	Open the config file.													 						*
 *----------------------------------------------------------------------------------*
 * returns: return value of Fopen()																	*
 *----------------------------------------------------------------------------------*
 *		name:			name of file to open from $HOME or local directory						*
 * 	mode:			same as by Fopen()																*
 * 	length:		will be filled out with the length of the file							*
 *==================================================================================*/
static long open_file(char *name, int mode, long *length)
{
	long ret = EFILNF;
	long xret;
	int len;
	char *env_value;
	char **env_result = &env_value;
	char filename[256] = "";
	XATTR xattr;

	shel_envrn(env_result, "HOME=");
	if (env_value != NULL)								// if $HOME exists
	{
		#ifdef debug_open_file
			printf("env_value = \"%s\"\r\n", env_value);
		#endif
		strcpy(filename, env_value);
		#ifdef debug_open_file
			printf("filename = \"%s\"\r\n", filename);
		#endif
		len = strlen(filename);
		#ifdef debug_open_file
			printf("len = %i\r\n", len);
		#endif
		if ((filename[len - 1] != '\\') && (filename[len - 1] != '/'))
		{
			strcat(filename, "\\");
		}
	}
	strcat(filename, name);
	#ifdef debug_open_file
		printf("filename + name = \"%s\"\r\n", filename);
	#endif
	if((xret = Fxattr(0, filename, &xattr)) != 0)	// if file doesn't exist in $HOME or even $HOME doesn't exist
	{
		#ifdef debug_open_file
			printf("look in current directory\r\n");
		#endif
		strcpy(filename, name);							// look for file in current directory
		xret = Fxattr(0, filename, &xattr);
	}
	if(xret == 0)
	{
		*length = xattr.size;
		if (mode == 1)		// S_WRITE
			ret = Fcreate(filename, 0x0020);
		else
			ret = Fopen(filename, mode);
	}
	return ret;
}

/*==================================================================================*
 * save_cnf:																								*
 * 	Save a state to config file.											 						*
 *----------------------------------------------------------------------------------*
 * returns: 0 if no error occured else 1.															*
 *----------------------------------------------------------------------------------*
 *==================================================================================*/
static int save_cnf(void)
{
	long ret;
	int cnf;
	long size;
	int error;
	char line[256];

	if ((ret = open_file(INFname, 1, &size)) < 0)
	{
		printf("\033H\033BKEYTBL: Error while opening \"%s\"\r\n", INFname);
		return 1;
	}
	cnf = 0x0000ffff & ret;
	error = 0;

	cnf_buffer[0] = '\0';
	if (sprintf(line, "language=%s\r\n", language[current_lang].suffix) < 0)
	{
		error = 1;
		goto save_cnf_abort;
	}
	strcat(cnf_buffer, line);
	if (sprintf(line, "right_offset=%d\r\n", right_offset) < 0)
	{
		error = 1;
		goto save_cnf_abort;
	}
	strcat(cnf_buffer, line);
	if (sprintf(line, "icon_set=%d\r\n", icon_set) < 0)
	{
		error = 1;
		goto save_cnf_abort;
	}
	strcat(cnf_buffer, line);
	if (sprintf(line, "swap_29_2b=%d\r\n", swap_29_2b) < 0)
	{
		error = 1;
		goto save_cnf_abort;
	}
	strcat(cnf_buffer, line);

//	printf("cnf_buffer=\r\n%s", cnf_buffer);
	size = strlen(cnf_buffer);
//	printf("size=%li\r\n", size);
	if (Fwrite(cnf, size, cnf_buffer) != size)
		error = 1;

save_cnf_abort:
	if (error)
	{
		printf("\033H\033BKEYTBL: Error while writing \"%s\"\r\n", INFname);
		return 1;
	}

	Fclose(cnf);

	return error;
}

/*==================================================================================*
 * parse_cnfbuffer:																						*
 * 	Parses the content of cnf_buffer and assigns the values to the						*
 * 	corresponding global variables.																*
 *----------------------------------------------------------------------------------*
 * returns: 0 if no error occured else 1.															*
 *----------------------------------------------------------------------------------*
 *==================================================================================*/
static int parse_cnfbuffer(void)
{
	char line[256];
	int cnf_pos, l_pos;
	int error;

	error = 0;
	cnf_pos = 0;
	while ((cnf_buffer[cnf_pos] != '\0') && (error == 0))
	{
		/* get next line */
		l_pos = 0;
		while ((cnf_buffer[cnf_pos] != '\n')
					&& (cnf_buffer[cnf_pos] != '\r')
					&& (cnf_buffer[cnf_pos] != '\0'))
			line[l_pos++] = cnf_buffer[cnf_pos++];
		line[l_pos] = '\0';	// string ending
		while ((cnf_buffer[cnf_pos] == '\n') || (cnf_buffer[cnf_pos] == '\r'))
			cnf_pos++;			// skip '\n' and '\r'

		if (sscanf(line, "language=%s", cnf_lang_s) == 1);
		else if (sscanf(line, "right_offset=%d", &right_offset) == 1);
		else if (sscanf(line, "icon_set=%d", &icon_set) == 1);
		else if (sscanf(line, "swap_29_2b=%d", &swap_29_2b) == 1);
		else
			error = 1;
	}
	return error;
}

/*==================================================================================*
 * load_cnf:																								*
 * 	Load configuration file.																		*
 *----------------------------------------------------------------------------------*
 * returns: 0 if no error occured else 1.															*
 *----------------------------------------------------------------------------------*
 *==================================================================================*/
static int load_cnf(void)
{
	long ret;
	int cnf;
	long size;

	if ((ret = open_file(INFname, 0, &size)) < 0)
	{
		printf("\033H\033BKEYTBL: Error while opening \"%s\"\r\n", INFname);
		return 1;
	}
	cnf = 0x0000ffff & ret;

	if (size >= CNF_BUFFER_LEN)
		goto load_cnf_err;
	if (Fread(cnf, size, &cnf_buffer) != size)
		goto load_cnf_err;
	cnf_buffer[size] = '\0';
	if (parse_cnfbuffer() == 1)
		goto load_cnf_err;

	Fclose(cnf);
	return 0;

load_cnf_err:
	Fclose(cnf);
	printf("\033H\033BKEYTBL: Error while reading \"%s\"\r\n", INFname);
	return 1;
}



/*************************************
   Handler procedures for user input
 *************************************/

static void handle_msg(int *msg)
{
	switch (msg[0])				/* msg[0] is message type */
	{
		case AP_TERM :
			quit = TRUE;
			break;
		case MN_SELECTED:
		case WM_REDRAW :
		case WM_NEWTOP :
		case WM_TOPPED :
		case WM_CLOSED :
		case WM_FULLED :
		case WM_SIZED :
		case WM_MOVED :
		case WM_BOTTOMED :
		case WM_ICONIFY :
		case WM_UNICONIFY :
			break;
		default:
			break;
	}
}



/********
   Main
 ********/

/*==================================================================================*
 * update_right_offset:																					*
 * 	Update positions of icon and popup menu according to value of the global		*
 *		variable right_offset.																			*
 *----------------------------------------------------------------------------------*
 * returns: -																								*
 *----------------------------------------------------------------------------------*
 *==================================================================================*/
static void update_right_offset(void)
{
	int lang, icnset;

	// position of flag icon area
	area.g_x = (right_offset >= 0 ? screen.g_w - FLAG_W : 0) - right_offset;
	area.g_y = top_offset;
	area.g_w = FLAG_W;
	area.g_h = FLAG_H;
//	printf("\033H\033Barea: %i %i %i %i\r\n", area.g_x, area.g_y, area.g_w, area.g_h);

	// position of flag icons+
	for (lang = 0; lang < noflangs; lang++)
	{
		for (icnset = 0; icnset < NOOFICONSETS; icnset++)
		{
			language[lang].tree[icnset][0].ob_x = area.g_x;			/* set coordinates */
			language[lang].tree[icnset][0].ob_y = area.g_y;			/* ... */
		}
	}

	// position of popup menu
	rsrc_gaddr(R_TREE, selection_ext, &tree);
	menu_x = (right_offset >= 0 ? desk.g_x + desk.g_w - FLAG_W : 0) - right_offset - tree[selection_text_ext].ob_x;
	if (menu_x < 1)
		menu_x = 1;
	if (menu_x + tree[selection_parent_ext].ob_width >= desk.g_x + desk.g_w - 1)
		menu_x = desk.g_x + desk.g_w - tree[selection_parent_ext].ob_width - 1;
	menu_y = desk.g_y + 1;
}

/* create languages popup with currently selected icon set */
static void build_popup(void)
{
	int entry_h, i, check_x, icon_x, icon_y, text_x, y;

	rsrc_gaddr(R_TREE, selection_ext, &tree);

	entry_h = tree[selection_text_ext].ob_height;
	check_x = tree[selection_check_ext].ob_x;
	icon_x = tree[selection_icon_ext].ob_x;
	icon_y = tree[selection_icon_ext].ob_y;
	text_x = tree[selection_text_ext].ob_x;
	if (noflangs > 1)
	{
		sel_tree[0] = tree[selection_parent_ext];
		sel_tree[0].ob_height = entry_h * (noflangs - 1);
		sel_tree[0].ob_next = -1;
		sel_tree[0].ob_head = 1;
		sel_tree[0].ob_tail = OBJ_PER_LANG * (noflangs - 1);
		y = 0;
		for (i = 1; i < noflangs; i++)
		{
			// string for check mark
			sel_tree[OBJ_PER_LANG * i - 2] = tree[selection_check_ext];
			sel_tree[OBJ_PER_LANG * i - 2].ob_next = OBJ_PER_LANG * i;
			sel_tree[OBJ_PER_LANG * i - 2].ob_head = OBJ_PER_LANG * i - 1;
			sel_tree[OBJ_PER_LANG * i - 2].ob_tail = OBJ_PER_LANG * i - 1;
			sel_tree[OBJ_PER_LANG * i - 2].ob_x = check_x;
			sel_tree[OBJ_PER_LANG * i - 2].ob_y = y;
	
			// flag icon
			sel_tree[OBJ_PER_LANG * i - 1] = language[i].tree[icon_set][1];
			sel_tree[OBJ_PER_LANG * i - 1].ob_next = OBJ_PER_LANG * i - 2;
			sel_tree[OBJ_PER_LANG * i - 1].ob_head = -1;
			sel_tree[OBJ_PER_LANG * i - 1].ob_tail = -1;
			sel_tree[OBJ_PER_LANG * i - 1].ob_x = icon_x;
			sel_tree[OBJ_PER_LANG * i - 1].ob_y = icon_y;
	
			// name of language
			sel_tree[OBJ_PER_LANG * i] = tree[selection_text_ext];
			if (i < noflangs - 1)
			{
				sel_tree[OBJ_PER_LANG * i].ob_next = OBJ_PER_LANG * i + 1;
				sel_tree[OBJ_PER_LANG * i].ob_flags &= ~LASTOB;			// clear LASTOB flag
			}
			else
			{
				sel_tree[OBJ_PER_LANG * i].ob_next = 0;
				sel_tree[OBJ_PER_LANG * i].ob_flags |= LASTOB;			// set LASTOB flag
			}
			sel_tree[OBJ_PER_LANG * i].ob_head = -1;
			sel_tree[OBJ_PER_LANG * i].ob_tail = -1;
			sel_tree[OBJ_PER_LANG * i].ob_x = text_x;
			sel_tree[OBJ_PER_LANG * i].ob_y = y;
			sel_tree[OBJ_PER_LANG * i].ob_spec.free_string = language[i].language;
	
			y += entry_h;
		}
	}
	else
	{
		sel_tree[0] = tree[selection_parent_ext];
		sel_tree[0].ob_height = entry_h;
		sel_tree[0].ob_next = -1;
		sel_tree[0].ob_head = 1;
		sel_tree[0].ob_tail = OBJ_PER_LANG;

		// string for check mark
		sel_tree[1] = tree[selection_check_ext];
		sel_tree[1].ob_next = 3;
		sel_tree[1].ob_flags &= ~SELECTABLE;	// make not selectable
		sel_tree[1].ob_head = 2;
		sel_tree[1].ob_tail = 2;
		sel_tree[1].ob_x = check_x;
		sel_tree[1].ob_y = 0;

		// flag icon
		sel_tree[2] = language[0].tree[icon_set][1];
		sel_tree[2].ob_next = 1;
		sel_tree[2].ob_head = -1;
		sel_tree[2].ob_tail = -1;
		sel_tree[2].ob_x = icon_x;
		sel_tree[2].ob_y = icon_y;

		// name of language
		sel_tree[3] = tree[selection_text_ext];
		sel_tree[3].ob_next = 0;
		sel_tree[3].ob_flags |= LASTOB;			// set LASTOB flag
		sel_tree[3].ob_flags &= ~SELECTABLE;	// make not selectable
		sel_tree[3].ob_head = -1;
		sel_tree[3].ob_tail = -1;
		sel_tree[3].ob_x = text_x;
		sel_tree[3].ob_y = 0;
		sel_tree[3].ob_spec.free_string = language[0].language;
	}
/* *** included in update_right_offset() beeing called later
	menu_x = (right_offset >= 0 ? desk.g_x + desk.g_w - FLAG_W : 0) - right_offset - tree[selection_text_ext].ob_x;
*/
/*
	if (menu_x < 1)
		menu_x = 1;
	if (menu_x + tree[selection_parent_ext].ob_width >= desk.g_x + desk.g_w - 1)
		menu_x = desk.g_x + desk.g_w - tree[selection_parent_ext].ob_width - 1;
*/
/*
	menu_y = desk.g_y + 1;
*/
}



/*==================================================================================*
 * swap_c:																									*
 *		Swaps two chars.																					*
 *----------------------------------------------------------------------------------*
 * returns: -																								*
 *----------------------------------------------------------------------------------*
 *		a:		pointer to first char																	*
 *		b:		pointer to second char																	*
 *==================================================================================*/
static void swap_c(char *a, char *b)
{
	char tmp;

	tmp = *a;
	*a = *b;
	*b = tmp;
}



/*==================================================================================*
 * swap_keys:																								*
 *		Swaps two keys in the current keyboard layout.											*
 *----------------------------------------------------------------------------------*
 * returns: -																								*
 *----------------------------------------------------------------------------------*
 *		keytab:	pointer to keyboard layout struct												*
 *		key1:		scan code of first key																*
 *		key2:		scan code of second key																*
 *==================================================================================*/
static void swap_keys(MGKEYTAB *keytab, int key1, int key2)
{
	swap_c(&keytab->unshift[key1], &keytab->unshift[key2]);
	swap_c(&keytab->shift[key1], &keytab->shift[key2]);
	swap_c(&keytab->caps[key1], &keytab->caps[key2]);
	swap_c(&keytab->altgr_unshift[key1], &keytab->altgr_unshift[key2]);
	swap_c(&keytab->altgr_shift[key1], &keytab->altgr_shift[key2]);
	swap_c(&keytab->altgr_caps[key1], &keytab->altgr_caps[key2]);
	swap_c(&keytab->alt_unshift[key1], &keytab->alt_unshift[key2]);
	swap_c(&keytab->alt_shift[key1], &keytab->alt_shift[key2]);
	swap_c(&keytab->alt_caps[key1], &keytab->alt_caps[key2]);
}



/* load keyboard layout of currently selected language */
static void set_layout(void)
{
	MGKEYTAB *keytab;
	char file[256];
	long ret;
	int fhdl;
//	long super_sp;

	keytab = Keytbl((char *)-1, (char *)-1, (char *)-1);
	#ifdef debug_keytbladdr
		printf("keytab->unshift =       %p\r\n", keytab->unshift);
		printf("keytab->shift =         %p\r\n", keytab->shift);
		printf("keytab->caps =          %p\r\n", keytab->caps);
		printf("keytab->altgr_unshift = %p\r\n", keytab->altgr_unshift);
		printf("keytab->altgr_shift =   %p\r\n", keytab->altgr_shift);
		printf("keytab->altgr_caps =    %p\r\n", keytab->altgr_caps);
		printf("keytab->alt_unshift =   %p\r\n", keytab->alt_unshift);
		printf("keytab->alt_shift =     %p\r\n", keytab->alt_shift);
		printf("keytab->alt_caps =      %p\r\n", keytab->alt_caps);
	#endif
	strcpy(file, keytabls_path);
	strcat(file, "\\");
	strcat(file, LAYOUTFILE);
	strcat(file, ".");
	strcat(file, language[current_lang].suffix);
	#ifdef debug_setlayout
		printf("file=\"%s\"\r\n", file);
	#endif
	if ((ret = Fopen(file, 0)) < 0)
	{
		#ifdef debug_setlayout
			printf("set_layout(): Error while Fopen()\r\n");
		#endif
		return;
	}
	fhdl = (int)ret;
	if ((ret = Fread(fhdl, KEYTABLSLEN, keytabls_buf)) < KEYTABLSLEN)
	{
		Fclose(fhdl);
		#ifdef debug_setlayout
			printf("set_layout(): Error while Fread()\r\n");
		#endif
		return;
	}
	Fclose(fhdl);
//	super_sp = Super((void *)0);		// go to supervisor mode
	memcpy(keytab->unshift, &keytabls_buf[KEYTAB_UNSHIFT], KEYTABLSKEYS);
	memcpy(keytab->shift, &keytabls_buf[KEYTAB_SHIFT], KEYTABLSKEYS);
	memcpy(keytab->caps, &keytabls_buf[KEYTAB_CAPS], KEYTABLSKEYS);
	memcpy(keytab->altgr_unshift, &keytabls_buf[KEYTAB_ALTGR_UNSHIFT], KEYTABLSKEYS);
	memcpy(keytab->altgr_shift, &keytabls_buf[KEYTAB_ALTGR_SHIFT], KEYTABLSKEYS);
	memcpy(keytab->altgr_caps, &keytabls_buf[KEYTAB_ALTGR_CAPS], KEYTABLSKEYS);
	memcpy(keytab->alt_unshift, &keytabls_buf[KEYTAB_ALT_UNSHIFT], KEYTABLSKEYS);
	memcpy(keytab->alt_shift, &keytabls_buf[KEYTAB_ALT_SHIFT], KEYTABLSKEYS);
	#ifdef keytabls_safecopy
		#define SAVESUB 16
	#else
		#define SAVESUB 0
	#endif
	memcpy(keytab->alt_caps, &keytabls_buf[KEYTAB_ALT_CAPS], KEYTABLSKEYS - SAVESUB);

	if (swap_29_2b != 0)
		swap_keys(keytab, 0x29, 0x2b);

//	Super((void *)super_sp);			// return to user mode
}



/* redraw the flag */
static void redraw_depth(int depth)
{
	OBJECT *curflag;

	if ((xssi_struct == NULL) || (xssi_struct->xssi_save_stat == 0))
	{
		curflag = &(language[current_lang].tree[icon_set][0]);
		wind_update(BEG_UPDATE);
//		wind_update(BEG_MCTRL);
		objc_draw(curflag, 0, depth, curflag[0].ob_x, curflag[0].ob_y, curflag[0].ob_width, curflag[0].ob_height);
//		wind_update(END_MCTRL);
		wind_update(END_UPDATE);
	}
}

static void redraw(void)
{
	redraw_depth(7);
}

static void redraw_background(void)
{
	redraw_depth(0);
}



/*==================================================================================*
 * move_position:																							*
 *		Moves position of the flag icon until all mouse buttons are released. A box	*
 *		which is moved with the mouse cursor indicates the position. The offsets		*
 *		from the left resp. right screen borders are displayed below the box.			*
 *		Updates the value of the variable "right_offset".
 *----------------------------------------------------------------------------------*
 * returns: 1 if final position is different from the initial, 0 else.					*
 *----------------------------------------------------------------------------------*
 *==================================================================================*/
static int move_position(void)
{
	int mx, my, mb, ks;
	int old_mx, x_off, cur_pos_x, prev_pos_x, initial_pos_x;
	GRECT rect;
	OBJECT *tree, *tree_left, *tree_right;
	int position_text, offset;
	int pxy[4];



	cur_pos_x = area.g_x;
	prev_pos_x = cur_pos_x;
	initial_pos_x = cur_pos_x;

	graf_mkstate(&mx, &my, &mb, &ks);
	old_mx = mx;
	x_off = mx - area.g_x;

	rsrc_gaddr(R_TREE, position_left_ext, &tree_left);
	rsrc_gaddr(R_TREE, position_right_ext, &tree_right);

	// display current position
	if (right_offset >= 0)
	{
		tree = tree_right;
		position_text = position_right_text_ext;
		offset = right_offset;
	}
	else
	{
		tree = tree_left;
		position_text = position_left_text_ext;
		offset = - right_offset;
	}
	sprintf(tree[position_text].ob_spec.tedinfo->te_ptext, "%i", offset);
	tree[ROOT].ob_x = right_offset >= 0 ? screen.g_w - right_offset - tree[ROOT].ob_width : - right_offset;
	if (tree[ROOT].ob_x + tree[ROOT].ob_width >= screen.g_w)
		tree[ROOT].ob_x = screen.g_w - tree[ROOT].ob_width;
	if (tree[ROOT].ob_x < 0)
		tree[ROOT].ob_x = 0;
	tree[ROOT].ob_y = desk.g_y;
	objc_rect(tree, ROOT, &rect);
	form_dial(FMD_START, 0, 0, 0, 0, rect.g_x, rect.g_y, rect.g_w, rect.g_h);
	objc_draw(tree, ROOT, 7, rect.g_x, rect.g_y, rect.g_w, rect.g_h);

	// set color an drawing mode parameters for box
	vsf_color(vdi_handle, WHITE);
	vsf_interior(vdi_handle, FIS_SOLID);
	vsf_perimeter(vdi_handle, PERIMETER_OFF);
	vswr_mode(vdi_handle, MD_XOR);

	// draw box
	graf_mouse(M_OFF, NULL);
	pxy[0] = cur_pos_x;
	pxy[1] = area.g_y;
	pxy[2] = pxy[0] + FLAG_W - 1;
	pxy[3] = pxy[1] + FLAG_H - 1;
	v_bar(vdi_handle, pxy);
	graf_mouse(M_ON, NULL);

	while (mb != 0)
	{
		if (mx != old_mx)
		{
			if ((old_mx < 1 + x_off) || (old_mx > screen.g_x + screen.g_w - FLAG_W + x_off))		// if mouse was in "dead" area
			{
				if ((mx >= 1 + x_off) && (mx <= screen.g_x + screen.g_w - FLAG_W + x_off))			// if mouse exited "dead" area
					cur_pos_x += mx - (prev_pos_x + x_off);
			}
			else
				cur_pos_x += mx - old_mx;

			if (cur_pos_x < 1)
				cur_pos_x = 1;
			else if (cur_pos_x > screen.g_x + screen.g_w - FLAG_W)
				cur_pos_x = screen.g_x + screen.g_w - FLAG_W;

			if (cur_pos_x != prev_pos_x)
			{
				right_offset = (cur_pos_x < SWITCH_OFFSET) ? -cur_pos_x : screen.g_x + screen.g_w - (cur_pos_x + FLAG_W);

				// move box to new position
				graf_mouse(M_OFF, NULL);
				pxy[0] = prev_pos_x;
				pxy[1] = area.g_y;
				pxy[2] = pxy[0] + FLAG_W - 1;
				pxy[3] = pxy[1] + FLAG_H - 1;
				v_bar(vdi_handle, pxy);
				pxy[0] = cur_pos_x;
				pxy[1] = area.g_y;
				pxy[2] = pxy[0] + FLAG_W - 1;
				pxy[3] = pxy[1] + FLAG_H - 1;
				v_bar(vdi_handle, pxy);
				graf_mouse(M_ON, NULL);

				prev_pos_x = cur_pos_x;

				// display current position
				form_dial(FMD_FINISH, 0, 0, 0, 0, rect.g_x, rect.g_y, rect.g_w, rect.g_h);
				if (right_offset >= 0)
				{
					tree = tree_right;
					position_text = position_right_text_ext;
					offset = right_offset;
				}
				else
				{
					tree = tree_left;
					position_text = position_left_text_ext;
					offset = - right_offset;
				}
				sprintf(tree[position_text].ob_spec.tedinfo->te_ptext, "%i", offset);
				tree[ROOT].ob_x = right_offset >= 0 ? screen.g_w - right_offset - tree[ROOT].ob_width : - right_offset;
				if (tree[ROOT].ob_x + tree[ROOT].ob_width >= screen.g_w)
					tree[ROOT].ob_x = screen.g_w - tree[ROOT].ob_width;
				if (tree[ROOT].ob_x < 0)
					tree[ROOT].ob_x = 0;
				tree[ROOT].ob_y = desk.g_y;
				objc_rect(tree, ROOT, &rect);
				form_dial(FMD_START, 0, 0, 0, 0, rect.g_x, rect.g_y, rect.g_w, rect.g_h);
				objc_draw(tree, ROOT, 7, rect.g_x, rect.g_y, rect.g_w, rect.g_h);
			}
		}
		old_mx = mx;
		graf_mkstate(&mx, &my, &mb, &ks);
	}

	// erase box
	graf_mouse(M_OFF, NULL);
	pxy[0] = cur_pos_x;
	pxy[1] = area.g_y;
	pxy[2] = pxy[0] + FLAG_W - 1;
	pxy[3] = pxy[1] + FLAG_H - 1;
	v_bar(vdi_handle, pxy);
	graf_mouse(M_ON, NULL);

	form_dial(FMD_FINISH, 0, 0, 0, 0, rect.g_x, rect.g_y, rect.g_w, rect.g_h);

	return cur_pos_x != initial_pos_x ? 1 : 0;

/*** excerpt from old version ***
	if (graf_dragbox(FLAG_W, FLAG_H, area.g_x, area.g_y, 1, top_offset, screen.g_w - 1, FLAG_H, &endx, &endy) != 0)
	{
		form_dial(FMD_FINISH, 0, 0, 0, 0, rect.g_x, rect.g_y, rect.g_w, rect.g_h);
		#ifdef debug_moveicon
			printf("endx=%i\r\n", endx);
			printf("endy=%i\r\n", endy);
		#endif
		redraw_background();		// erase flag on old position
		right_offset = (endx < 100) ? -endx : screen.g_x + screen.g_w - (endx + FLAG_W);
		update_right_offset();	// draw flag on new position
		redraw();

		// display new position
		if (right_offset >= 0)
		{
			position = position_right_ext;
			position_text = position_right_text_ext;
			offset = right_offset;
		}
		else
		{
			position = position_left_ext;
			position_text = position_left_text_ext;
			offset = - right_offset;
		}
		rsrc_gaddr(R_TREE, position, &tree);
		sprintf(tree[position_text].ob_spec.tedinfo->te_ptext, "%i", offset);
		tree[ROOT].ob_x = right_offset >= 0 ? screen.g_w - right_offset - tree[ROOT].ob_width : - right_offset;
		if (tree[ROOT].ob_x + tree[ROOT].ob_width >= screen.g_w)
			tree[ROOT].ob_x = screen.g_w - tree[ROOT].ob_width;
		if (tree[ROOT].ob_x < 0)
			tree[ROOT].ob_x = 0;
		tree[ROOT].ob_y = desk.g_y;
		objc_rect(tree, ROOT, &rect);
		form_dial(FMD_START, 0, 0, 0, 0, rect.g_x, rect.g_y, rect.g_w, rect.g_h);
		objc_draw(tree, ROOT, 7, rect.g_x, rect.g_y, rect.g_w, rect.g_h);
		time_start = time_current = get_hz200();
		while (time_current - time_start < POSITION_SHOW_HZ200)
			time_current = get_hz200();
		form_dial(FMD_FINISH, 0, 0, 0, 0, rect.g_x, rect.g_y, rect.g_w, rect.g_h);
	}
*/
}



/* returns 1 if language "lang" (suffix) is in array "layout", else 0 */
static int lang_avail(char *lang)
{
	int i;
	for (i = 0; i < noflayouts; i++)
		if (strcmp(layout[i], lang) == 0)
			return 1;
	return 0;
}



int screen_open(void)
{
	int i;

	vdi_handle = graf_handle(&wcell, &hcell, &wbox, &hbox);
	work_in[0] = Getrez() + 2;
	for (i = 1; i < 10; work_in[i++] = 1);
	work_in[10] = 2;
	v_opnvwk(work_in, &vdi_handle, work_out);
	return (vdi_handle);
}



int main(int argc, char *argv[])
{
	char *altNoVDIWork = "[3][GEM is unable to|allocate a workstation.|The program must abort.][ OK ]";
	char *altNoRSC = "[3][The program cannot locate|KEYTBL.RSC. Pleas ensure|that it resides in the|same directory as|KEYTBL.APP.][ OK ]";

	int	msg[8];						/* Message buffer */
	int	events;						/* What events are valid? */	
	int	msx, msy;					/* Mouse position */
	int	mbutton, mclick;			/* Mouse buttons/#clicks */
	int	kstate, kreturn;			/* Key state/code */
	int	bclicks, bmask, bstate;

	int	obj, icnset;
	long	dirhandle;

	ap_id = appl_init();
	if(ap_id == -1)
		return -1;


	if (screen_open() == 0)
	{
		graf_mouse(ARROW, 0L);
		form_alert(1, altNoVDIWork);
		appl_exit();
		return -1;
	}

	if (!rsrc_load(RSCname))
	{
		graf_mouse(ARROW, 0L);
		form_alert(1, altNoRSC);
		v_clsvwk(vdi_handle);
		appl_exit();
		return -1;
	}

	/* get screen and desktop sizes */
	wind_get_grect(0, WF_WORKXYWH, &desk);
	wind_get_grect(0, WF_CURRXYWH, &screen);

	sm_id = appl_find("SCRENMGR");

	/* look for xssi cookie */
	get_cookie_s("XSSI", (long *)&xssi_struct);
/*
	if (get_cookie_s("XSSI", (long *)&xssi_struct))
	{
		printf("XSSI cookie found:\r\n");
		printf("xssi_id = %s\r\n", (char *)&(xssi_struct->xssi_id));
		printf("xssi_version = %x\r\n", xssi_struct->xssi_version);
		printf("xssi_save_stat = %x\r\n", xssi_struct->xssi_save_stat);
		printf("xssi_prg_stat = %x\r\n", xssi_struct->xssi_prg_stat);
		printf("xssi_vec_stat = %lx\r\n", xssi_struct->xssi_vec_stat);
	}
*/

	/* load config file */
	if (load_cnf() == 1)
	{
		goto exit;
	}
	#ifdef debug_getlangs
		printf("cnf_lang_s=\"%s\"\r\n", cnf_lang_s);
	#endif

	/* set area */
/* *** included in update_right_offset() beeing called later
	area.g_x = (right_offset >= 0 ? screen.g_w - FLAG_W : 0) - right_offset;
	area.g_y = top_offset;
	area.g_w = FLAG_W;
	area.g_h = FLAG_H;
//	printf("\033H\033Barea: %i %i %i %i\r\n", area.g_x, area.g_y, area.g_w, area.g_h);
*/

	/* get available languages */
	/* get available keyboard layouts */
	noflayouts = 0;
	if (((dirhandle = Dopendir(keytabls_path, 0)) & 0xff000000) == 0xff000000)
	{
		printf("Error: Dopendir() returned %li\r\n", dirhandle);
		goto exit;
	}
	{
		char buf[DREADDIRBUFLEN];
		long ret;
		#ifdef debug_getlangs
			printf("Dreaddir()\r\n");
		#endif
		do {
			ret = Dreaddir(DREADDIRBUFLEN, dirhandle, buf);
			if ((ret == 0) && (noflayouts < MAXNOOFLANG))
			{
				#ifdef debug_getlangs
					printf("\"%s\"", &buf[4]);
				#endif
				str2upper(&buf[4]);		// transform filename to upper case
				#ifdef debug_getlangs
					printf(" -> str2upper() -> \"%s\"\r\n", &buf[4]);
				#endif
				if (strncmp(LAYOUTFILE, &buf[4], LAYOUTFILENAMELEN) == 0)
				{
					strcpy(layout[noflayouts++], &buf[4 + FILESUFFIXOFFSET]);
					#ifdef debug_getlangs
						printf("\"%s\" - \"%s\"\r\n", &buf[4], layout[noflayouts - 1]);
					#endif
				}
			}
		} while (ret == 0);
		#ifdef debug_getlangs
			printf("noflayouts=%i\r\n", noflayouts);
		#endif
		Dclosedir(dirhandle);
	}

	/* get available languages from ressource file */
	rsrc_gaddr(R_TREE, languages_ext, &tree);
	for (icnset = 0; icnset < NOOFICONSETS; icnset++)
	{
		language[0].tree[icnset][0] = tree[languages_iconbox_ext];
		language[0].tree[icnset][0].ob_next = -1;
		language[0].tree[icnset][0].ob_head = 1;
		language[0].tree[icnset][0].ob_tail = 1;
//		language[0].tree[icnset][0].ob_x = area.g_x;			/* set coordinates */
//		language[0].tree[icnset][0].ob_y = area.g_y;			/* ... */
		language[0].tree[icnset][1] = tree[languages_sys_ext + icnset];
		language[0].tree[icnset][1].ob_next = 0;
		language[0].tree[icnset][1].ob_head = -1;
		language[0].tree[icnset][1].ob_tail = -1;
		language[0].tree[icnset][1].ob_flags |= LASTOB;		/* make it last (and only) object in tree */
		language[0].tree[icnset][1].ob_x = 0;					/* set coordinates */
		language[0].tree[icnset][1].ob_y = 0;					/* ... */
	}
	language[0].suffix = tree[languages_sys_ext + POS_SUFFIX].ob_spec.free_string;
	language[0].language = tree[languages_sys_ext + POS_LANGUAGE].ob_spec.free_string;
	noflangs = 1;
	obj = languages_prefirst_ext + 1;
	while (obj < languages_postlast_ext)
	{
		#ifdef debug_getlangs
			printf("\"%s\" - \"%s\"", tree[obj + POS_SUFFIX].ob_spec.free_string, tree[obj + POS_LANGUAGE].ob_spec.free_string);
		#endif
		if ((noflangs < MAXNOOFLANG) && (lang_avail(tree[obj + POS_SUFFIX].ob_spec.free_string)))
		{ /* max number of languages not yet reached and this language is available */
			#ifdef debug_getlangs
				printf(" > available ");
			#endif
			for (icnset = 0; icnset < NOOFICONSETS; icnset++)
			{
				language[noflangs].tree[icnset][0] = tree[languages_iconbox_ext];
				language[noflangs].tree[icnset][0].ob_next = -1;
				language[noflangs].tree[icnset][0].ob_head = 1;
				language[noflangs].tree[icnset][0].ob_tail = 1;
//				language[noflangs].tree[icnset][0].ob_x = area.g_x;			/* set coordinates */
//				language[noflangs].tree[icnset][0].ob_y = area.g_y;			/* ... */
				language[noflangs].tree[icnset][1] = tree[obj + icnset];
				language[noflangs].tree[icnset][1].ob_next = 0;
				language[noflangs].tree[icnset][1].ob_head = -1;
				language[noflangs].tree[icnset][1].ob_tail = -1;
				language[noflangs].tree[icnset][1].ob_flags |= LASTOB;		/* make it last (and only) object in tree */
				language[noflangs].tree[icnset][1].ob_x = 0;						/* set coordinates */
				language[noflangs].tree[icnset][1].ob_y = 0;						/* ... */
			}
			language[noflangs].suffix = tree[obj + POS_SUFFIX].ob_spec.free_string;
			language[noflangs].language = tree[obj + POS_LANGUAGE].ob_spec.free_string;
			if (strcmp(language[noflangs].suffix, cnf_lang_s) == 0)
				current_lang = noflangs;
			#ifdef debug_getlangs
				printf("\"%s\" - \"%s\"", language[noflangs].suffix, language[noflangs].language);
			#endif
			noflangs++;
		}
		#ifdef debug_getlangs
			printf("\r\n");
		#endif		
		obj += NOFOBJPERLINE;
	}
	#ifdef debug_getlangs
		printf("noflangs=%i\r\n", noflangs);
	#endif

	/* create popup for language selection */
	build_popup();
	update_right_offset();

	/* get number of bitplanes and adapt some colors if needed */
	{
		int work_out[57];

		vq_extnd(vdi_handle, 1, work_out);
		if (work_out[4] == 1)	// if black/white mode
		{
			rsrc_gaddr(R_TREE, position_left_ext, &tree);
			tree[position_left_box_ext].ob_spec.obspec.interiorcol = WHITE;
			rsrc_gaddr(R_TREE, position_right_ext, &tree);
			tree[position_right_box_ext].ob_spec.obspec.interiorcol = WHITE;
		}
	}

	/* set keyboard layout */
	if (current_lang != 0)
	{
		Bioskeys();
		set_layout();
	}


	graf_mouse(ARROW, 0L);

	quit = FALSE;
	while (!quit)
	{
		mbutton = 0;
		events = evnt_multi(MU_MESAG | MU_TIMER,
										2, 1, 1,
										0, 0, 0, 0, 0,
										0, 0, 0, 0, 0,
										msg,
										INTERVAL,
										&msx, &msy, &mbutton, &kstate,
										&kreturn, &mclick);

		if (events & MU_MESAG)
		{
			handle_msg(msg);
		}

		if (events & MU_TIMER)
		{
			int mx, my, mb, ks;
			graf_mkstate(&mx, &my, &mb, &ks);
			if (is_inside(mx, my, area))
			{
//				SysBell;
				wind_update(BEG_UPDATE);
				wind_update(BEG_MCTRL);

/*
				mbutton = 0;
				// wait for left click or mouse leaving the flag area
				events = evnt_multi(MU_BUTTON | MU_M1,
												2, 1, 1,
												1, area.g_x, area.g_y, area.g_w, area.g_h,
												0, 0, 0, 0, 0,
												msg,
												0,
												&msx, &msy, &mbutton, &kstate,
												&kreturn, &mclick);
*/

				mbutton = 0;
				// wait for left or right click or mouse leaving the flag area
				graf_mkstate(&msx, &msy, &mbutton, &kstate);
				if (!(mbutton & 3))
				{
					bclicks = 258;
					bmask = 3;
					bstate = 0;
				}
				else
				{
					bclicks = 0;
					bmask = 3;
					bstate = 0;
				}
				events = evnt_multi(MU_BUTTON | MU_M1,
												bclicks, bmask, bstate,
												1, area.g_x, area.g_y, area.g_w, area.g_h,
												0, 0, 0, 0, 0,
												msg,
												0,
												&msx, &msy, &mbutton, &kstate,
												&kreturn, &mclick);

//				wind_update(END_MCTRL);
//				wind_update(END_UPDATE);

				if (events & MU_BUTTON)
				{
//					SysBell;
					if (mbutton & 0x01)		/* left mouse button depressed */
					{
						if (kstate & K_ALT)	/* Alt depressed: change position of icon */
						{
							if (move_position() == 1)
							{
								redraw_background();		// erase flag on old position
								update_right_offset();	// update sll relevant positions
								redraw();					// draw flag on new position

/*
								// send redraw message to screen manager
								if (sm_id >= 0)			// if screen manager was found
								{
									short sm_msg[8];
	
									sm_msg[0] = 101;
									sm_msg[1] = ap_id;
									sm_msg[2] = 0;
									sm_msg[3] = 0;
									sm_msg[4] = 'MA';
									sm_msg[5] = 'GX';
									sm_msg[6] = 0;
									sm_msg[7] = 0;
									appl_write(sm_id, 16, sm_msg);
								}
*/
								save_cnf();
							}
						}
						else						/* show language popup */
						{
							int ret, old_lang;
							int i;
	
							for (i = 1; i < noflangs; i++)
							{
								sel_tree[OBJ_PER_LANG * i - 2].ob_state &= ~CHECKED;
							}
							if (current_lang > 0)
								sel_tree[OBJ_PER_LANG * current_lang - 2].ob_state |= CHECKED;
							old_lang = current_lang;
							if ((ret = lang_popup(sel_tree, 0, menu_x, menu_y)) >= 1)
							{
								if (ret % 3 == 0)
									current_lang = ret / OBJ_PER_LANG;
								else if (ret % 3 == 1)
									current_lang = (ret + 2) / OBJ_PER_LANG;
								if (current_lang != old_lang)
								{
									redraw();
									set_layout();
									save_cnf();
								}
							}
							#ifdef debug_langsel
								printf("ret = %i\r\n", ret);
								printf("current_lang = %i\r\n", current_lang);
							#endif
						}
					}
					if (mbutton & 0x02)		/* right mouse button depressed: cycle icon_set */
					{
						/* *** <-:|:-> RIGHT MOUSE BUTTON <-:|:-> *** */

						switch (icon_set)
						{
							case NOOFICONSETS - 1 :
								icon_set = 0;
								break;
							default:
								icon_set++;
						}
						build_popup();
						graf_mouse(M_OFF, NULL);
						redraw();
						graf_mouse(M_ON, NULL);
						save_cnf();
					}
				}

				wind_update(END_MCTRL);
				wind_update(END_UPDATE);
//				SysBell;
			}
			if (++interval_cnt > REDRAWCNT)
			{
				interval_cnt = 0;
				redraw();
			}
		}

	}


exit:
	v_clsvwk(vdi_handle);
	rsrc_free();
	appl_exit();
	return 0;
}
