/**********
   main.h
 **********/

// Constants

#define RSCname "keytbl.rsc"
#define INFname "keytbl.inf"

#define KEYTABLSLEN 1152				/* length of a "KEYTABLS.xxx" file */
#define KEYTABLSKEYS 128				/* number of key per keytabls sublayout */
#define KEYTAB_UNSHIFT (0 * KEYTABLSKEYS)
#define KEYTAB_SHIFT (1 * KEYTABLSKEYS)
#define KEYTAB_CAPS (2 * KEYTABLSKEYS)
#define KEYTAB_ALTGR_UNSHIFT (3 * KEYTABLSKEYS)
#define KEYTAB_ALTGR_SHIFT (4 * KEYTABLSKEYS)
#define KEYTAB_ALTGR_CAPS (5 * KEYTABLSKEYS)
#define KEYTAB_ALT_UNSHIFT (6 * KEYTABLSKEYS)
#define KEYTAB_ALT_SHIFT (7 * KEYTABLSKEYS)
#define KEYTAB_ALT_CAPS (8 * KEYTABLSKEYS)

#define INTERVAL 100						/* evnt_timer() interval in ms */
#define REDRAWCNT 20						/* redraw after this number of timer events */
#define CNF_BUFFER_LEN 128

#define OBJ_PER_LANG 3					/* number of objects for an entry in the language popup */
#define FLAG_W 16							/* width of flag icons */
#define FLAG_H 16							/* height of flag icons */

#define NOOFICONSETS 2					/* number of available icon sets */

#define POS_SUFFIX 2						/* position of suffix in RSC line */
#define POS_LANGUAGE 3					/* position of language name in RSC line */
#define NOFOBJPERLINE 4					/* number of objects per language description line in RSC */

#define MAXNOOFLANG 16					/* max number of languages supported */

#define DREADDIRBUFLEN 64				/* length of buffer for Dreaddir() */
#define LAYOUTFILE "KEYTABLS"			/* name of keyboard layout file without suffix */
#define LAYOUTFILENAMELEN 8			/*	length of string definde by LAYOUTFILE */
#define FILESUFFIXOFFSET 9				/* offset in filename "KEYTABLS.xxx" where suffix "xxx" begins */

#define POSITION_SHOW_HZ200 300		/* number of 1/200 seconds that position is displayed after changing position of flag icon */

#define SWITCH_OFFSET 100				/* positions of the flag left of this x-coordinate are taken relative to left screen border */



// Types

typedef struct
{
	unsigned short mode;
	long index;
	unsigned short dev;
	unsigned short reserved1;
	unsigned short nlink;
	unsigned short uid;
	unsigned short gid;
	long size;
	long blksize;
	long nblocks;
	short mtime;
	short mdate;
	short atime;
	short adate;
	short ctime;
	short cdate;
	short attr;
	short reserved2;
	long reserved3;
	long reserved4;
} XATTR;

typedef struct {
	OBJECT tree[NOOFICONSETS][2];		/* OBJECT trees containing flag icon of every icon set */
	char *suffix;							/* pointer to filename suffix for KEYTABLS.* file (max. 3 char + '\0') */
	char *language;						/* pointer to name of keyboard layout */
} LANGUAGE;

typedef struct {
	char *unshift;
	char *shift;
	char *caps;
	char *altgr_unshift;
	char *altgr_shift;
	char *altgr_caps;
	char *alt_unshift;
	char *alt_shift;
	char *alt_caps;
} MGKEYTAB;

typedef struct {
	long	xssi_id;
	short	xssi_version;
	short	xssi_save_stat;
	short	xssi_prg_stat;
	long	xssi_vec_stat;
} INFOXSSI;



// Service Procedures

static int is_inside(int x, int y, GRECT g);
static void objc_rect(OBJECT *tree, int obj, GRECT *rect);
static int lang_popup(OBJECT *tree, int obj, int x_off, int y_off);



// Config file functions

static long open_file(char *name, int mode, long *length);
static int save_cnf(void);
static int parse_cnfbuffer(void);
static int load_cnf(void);



// Handler procedures for user input

static void handle_msg(int *msg);



// Main

static void update_right_offset(void);
static void build_popup(void);
static void swap_c(char *a, char *b);
static void swap_keys(MGKEYTAB *keytab, int key1, int key2);
static void set_layout(void);
static void redraw_depth(int depth);
static void redraw(void);
static void redraw_background(void);
static int move_position(void);

static int lang_avail(char *lang);
int main(int argc, char* argv[]);
